@@ -24,6 +24,7 @@ use fields (
 	'outgoing_leg',   # Leg for outgoing_proxy
 	'queue',          # \@list of outstanding Net::SIP::Dispatcher::Packet
 	'response_cache', # Cache of responses, used to reply to retransmits
+	'disp_expire',    # expire/retransmit timer
 );
 
 use Net::SIP::Leg;
@@ -34,6 +35,7 @@ use List::Util 'first';
 use Net::DNS;
 use Carp 'croak';
 use Net::SIP::Debug;
+use Scalar::Util 'weaken';
 
 
 ###########################################################################
@@ -101,10 +103,16 @@ sub new {
 
 	# regularly prune queue
 	my $sub = sub {
-		my ($self,$loop) = @_;
-		$self->queue_expire($loop->looptime);
+		my ($self,$timer) = @_;
+		if ( $self ) {
+			$self->queue_expire( $self->{eventloop}->looptime );
+		} else {
+			$timer->cancel;
+		}
 	};
-	$self->add_timer( 1,[ $sub,$self,$eventloop ],1,'disp_expire' );
+	my $cb = [ $sub,$self ];
+	weaken( $cb->[1] );
+	$self->{disp_expire} = $self->add_timer( 1,$cb,1,'disp_expire' );
 
 	return $self;
 }
@@ -115,16 +123,22 @@ sub new {
 # Args: ($self,$receiver)
 #   $receiver: object which has receive( Net::SIP::Leg,Net::SIP::Packet )
 #     method to handle incoming SIP packets or callback
+#     might be undef - in this case the existing receiver will be removed
 # Returns: NONE
 ###########################################################################
 sub set_receiver {
 	my Net::SIP::Dispatcher $self = shift;
-	my $receiver = shift;
-	if ( my $sub = UNIVERSAL::can($receiver,'receive' )) {
-		# Object with method receive()
-		$receiver = [ $sub,$receiver ]
+	if ( my $receiver = shift ) {
+		if ( my $sub = UNIVERSAL::can($receiver,'receive' )) {
+			# Object with method receive()
+			$receiver = [ $sub,$receiver ]
+		}
+		$self->{receiver} = $receiver;
+	} else {
+		# remove receiver
+		$self->{receiver} = undef
 	}
-	$self->{receiver} = $receiver;
+		
 }
 
 ###########################################################################
@@ -162,6 +176,7 @@ sub add_leg {
 		if ( my $fd = $leg->fd ) {
 			my $cb = sub {
 				my ($self,$leg) = @_;
+				$self || return;
 
 				# leg->receive might return undef if the packet wasnt
 				# read successfully. for tcp connections the receive
@@ -176,7 +191,9 @@ sub add_leg {
 				# handle received packet
 				$self->receive( $packet,$leg,$from );
 			};
-			$self->{eventloop}->addFD( $fd, [ $cb,$self,$leg ]);
+			$cb = [ $cb,$self,$leg ];
+			weaken( $cb->[1] );
+			$self->{eventloop}->addFD( $fd, $cb );
 		}
 	}
 }
@@ -298,22 +315,45 @@ sub deliver {
 
 ###########################################################################
 # cancel delivery of all packets with specific id
-# Args: ($self,$id)
+# Args: ($self,$typ?,$id)
+#   $typ: what to cancel, e.g. 'id','callid' or 'qentry', optional,
+#     defaults to 'id' if $id is not ref or 'qentry' if $id is ref
 #   $id: id to cancel, can also be queue entry
 # Returns: NONE
 ###########################################################################
 sub cancel_delivery {
 	my Net::SIP::Dispatcher $self = shift;
-	my ($id) = @_;
-	my $q = $self->{queue};
-	if ( ref($id)) {
-		# it's a *::Dispatcher::Packet
-		DEBUG( 100,"cancel packet $id: $id->{id}" );
-		@$q = grep { $_ != $id } @$q;
+	my ($callid,$id,$qentry);
+	if ( @_ == 2 ) {
+		my $typ = shift;
+		if ( $typ eq 'callid' ) { $callid = shift }
+		elsif ( $typ eq 'id' ) { $id = shift }
+		elsif ( $typ eq 'qentry' ) { $qentry = shift }
+		else {
+			croak( "bad typ '$typ', should be id|callid|qentry" );
+		}
 	} else {
+		$id = shift;
+		if ( ref($id)) {
+			$qentry = $id;
+			$id = undef;
+		}
+	}
+	my $q = $self->{queue};
+	if ( $qentry ) {
+		# it's a *::Dispatcher::Packet
+		DEBUG( 100,"cancel packet id: $qentry->{id}" );
+		@$q = grep { $_ != $qentry } @$q;
+	} elsif ( defined $id ) {
 		no warnings; # $_->{id} can be undef
-		DEBUG( 100, "cancel packet $id" );
+		DEBUG( 100, "cancel packet id $id" );
 		@$q = grep { $_->{id} ne $id } @$q;
+	} elsif ( defined $callid ) {
+		no warnings; # $_->{callid} can be undef
+		DEBUG( 100, "cancel packet callid $callid" );
+		@$q = grep { $_->{callid} ne $callid } @$q;
+	} else {
+		croak( "cancel_delivery w/o id" );
 	}
 }
 
@@ -438,6 +478,9 @@ sub __deliver {
 
 	# loop until leg und dst_addr are known, when we call leg->deliver
 	my $leg = $qentry->{leg}[0];
+	if ( $leg && @{ $qentry->{leg}}>1 ) {
+		DEBUG( 50,"picking first of multiple legs: ".join( " ", map { $_->dump } @{ $qentry->{leg}} ));
+	}
 	my $dst_addr = $qentry->{dst_addr}[0];
 
 	if ( ! $dst_addr || ! $leg) {
@@ -476,6 +519,7 @@ sub __deliver {
 	# I have leg and addr, send packet thru leg to addr
 	my $cb = sub {
 		my ($self,$qentry,$error) = @_;
+		$self || return;
 		if ( !$error  && $qentry->{retransmits} ) {
 			# remove from queue even if timeout
 			$self->cancel_delivery( $qentry );
@@ -485,8 +529,11 @@ sub __deliver {
 
 	# adds via on cloned packet, calls cb if definite success (tcp)
 	# or error
-	DEBUG( 50,"deliver through leg $leg \@$dst_addr" );
-	$leg->deliver( $qentry->{packet},$dst_addr, [ $cb,$self,$qentry ] );
+	DEBUG( 50,"deliver through leg ".$leg->dump." \@$dst_addr" );
+	weaken( my $rself = \$self );
+	$cb = [ $cb,$self,$qentry ];
+	weaken( $cb->[1] );
+	$leg->deliver( $qentry->{packet},$dst_addr,$cb );
 
 	if ( !$qentry->{retransmits} ) {
 		# remove from queue if no timeout
@@ -606,6 +653,12 @@ sub resolve_uri {
 		@$dst_addr = ( $ip_addr );
 	}
 
+	# is param maddr set?
+	if ( my $ip = $param->{maddr} ) {
+		@$dst_addr = ( $ip ) 	
+			if $ip =~m{^[\d\.]+$} && eval { inet_aton($ip) };
+	}
+
 	# entries in form [ prio,proto,ip,port ]
 	my @resp;
 	foreach my $addr ( @$dst_addr ) {
@@ -620,6 +673,11 @@ sub resolve_uri {
 		}
 	}
 
+	# should we use a fixed transport?
+	if ( my $proto = $param->{transport} ) {
+		@resp = grep { lc($_->[1]) eq lc($proto) } @resp;
+	}
+
 	my @param = ( $dst_addr,$legs,$allowed_legs,$default_port,$callback );
 	return __resolve_uri_final( @param,0,\@resp ) if @resp;
 
@@ -731,16 +789,42 @@ sub dns_domain2srv {
 	my $dns = Net::DNS::Resolver->new;
 
 	# Try to get SRV records for _sip._udp.domain or _sip._tcp.domain
-	my @resp;
+	my (@resp,%addr2ip);
 	foreach my $proto ( @$protos ) {
 		if ( my $q = $dns->query( '_'.$sip_proto.'._'.$proto.'.'.$domain,'SRV' )) {
 			foreach my $rr ( $q->answer ) {
-				$rr->type eq 'SRV' || next;
-				# XXX fixme, get IPs for name
-				push @resp,[ $rr->priority, $proto,$rr->name,$rr->port ]
+				if ( $rr->type eq 'A' ) {
+					push @{ $addr2ip{$rr->name} }, $rr->address;
+				} elsif ( $rr->type eq 'SRV' ) {
+					push @resp,[ $rr->priority, $proto,$rr->target,$rr->port ]
+				}
 			}
 		}
 	}
+
+	# name to addr based on additional records in DNS answer
+	my @resp_resolved;
+	for my $r (@resp) {
+		if ( my $addr = $addr2ip{ $r->[2] } ) {
+			for (@$addr) {
+				my @cp = @$r;
+				$cp[2] = $_;
+				push @resp_resolved, \@cp;
+			}
+		} else {
+			# either already IP or no additional data for resolving -> later
+			my @cp = @$r;
+			# XXX fixme blocking DNS lookup
+			my $ipn = gethostbyname( $r->[2] ) or do {
+				DEBUG( 1,"cannot resolve $r->[2]" );
+				next;
+			};
+			$cp[2] = inet_ntoa($ipn);
+			push @resp_resolved, \@cp;
+		}
+	}
+	@resp = @resp_resolved;
+
 	# if no SRV records try to resolve address directly
 	unless (@resp) {
 		# try addr directly
@@ -748,7 +832,8 @@ sub dns_domain2srv {
 		if ( my $q = $dns->query( $domain,'A' )) {
 			foreach my $rr ($q->answer ) {
 				$rr->type eq 'A' || next;
-				# XXX fixme, get *all* IPs for name
+				# XXX fixme, check that name in response corresponds to query
+				# (beware of CNAMEs!)
 				push @resp,map {
 					[ -1, $_ , $rr->address,$default_port ]
 				} @$protos;
@@ -766,6 +851,7 @@ sub dns_domain2srv {
 package Net::SIP::Dispatcher::Packet;
 use fields (
 	'id',           # transaction id, used for canceling delivery if response came in
+	'callid',       # callid, used for canceling all deliveries for this call
 	'packet',       # the packet which nees to be delivered
 	'dst_addr',     # to which adress the packet gets delivered, is array-ref because
 					# the DNS/SRV lookup might return multiple addresses and protocols
@@ -796,6 +882,7 @@ sub new {
 	my $self = fields::new( $class );
 	%$self = %args;
 	$self->{id} ||= $self->{packet}->tid;
+	$self->{callid} ||= $self->{packet}->callid;
 
 	if ( my $addr = $self->{dst_addr} ) {
 		$self->{dst_addr} = [ $addr ] if !ref($addr)
